from dataclasses import field, dataclass

from kirin import ir
from kirin.passes import Pass
from bloqade.noise import native
from kirin.rewrite import (
    Walk,
    Chain,
    Fixpoint,
    ConstantFold,
    DeadCodeElimination,
    CommonSubexpressionElimination,
)
from bloqade.analysis import address
from kirin.rewrite.result import RewriteResult
from bloqade.qasm2.rewrite.heuristic_noise import NoiseRewriteRule


@dataclass
class NoisePass(Pass):
    """Apply a noise model to a quantum circuit.

    NOTE: This pass is not guaranteed to be supported long-term in bloqade. We will be
    moving towards a more general approach to noise modeling in the future.

    """

    noise_model: native.MoveNoiseModelABC = field(
        default_factory=native.TwoRowZoneModel
    )
    gate_noise_params: native.GateNoiseParams = field(
        default_factory=native.GateNoiseParams
    )
    address_analysis: address.AddressAnalysis = field(init=False)

    def __post_init__(self):
        self.address_analysis = address.AddressAnalysis(self.dialects)

    def unsafe_run(self, mt: ir.Method):
        result = RewriteResult()

        frame, res = self.address_analysis.run_analysis(mt, no_raise=False)
        result = (
            Walk(
                NoiseRewriteRule(
                    address_analysis=frame.entries,
                    noise_model=self.noise_model,
                    gate_noise_params=self.gate_noise_params,
                )
            )
            .rewrite(mt.code)
            .join(result)
        )
        rule = Chain(
            ConstantFold(),
            DeadCodeElimination(),
            CommonSubexpressionElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)
        return result



from typing import Dict, List, Tuple
from dataclasses import field, dataclass

from kirin import ir
from bloqade.noise import native
from kirin.rewrite import abc as result_abc, result
from kirin.dialects import py, ilist
from bloqade.analysis import address
from bloqade.qasm2.dialects import uop, core, glob, parallel


@dataclass
class NoiseRewriteRule(result_abc.RewriteRule):
    """
    NOTE: This pass is not guaranteed to be supported long-term in bloqade. We will be
    moving towards a more general approach to noise modeling in the future.
    """

    address_analysis: Dict[ir.SSAValue, address.Address]
    gate_noise_params: native.GateNoiseParams = field(
        default_factory=native.GateNoiseParams
    )
    noise_model: native.MoveNoiseModelABC = field(
        default_factory=native.TwoRowZoneModel
    )
    qubit_ssa_value: Dict[int, ir.SSAValue] = field(default_factory=dict, init=False)

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if isinstance(node, core.QRegNew):
            return self.rewrite_qreg_new(node)
        elif isinstance(node, uop.SingleQubitGate):
            return self.rewrite_single_qubit_gate(node)
        elif isinstance(node, uop.CZ):
            return self.rewrite_cz_gate(node)
        elif isinstance(node, (parallel.UGate, parallel.RZ)):
            return self.rewrite_parallel_single_qubit_gate(node)
        elif isinstance(node, parallel.CZ):
            return self.rewrite_parallel_cz_gate(node)
        elif isinstance(node, glob.UGate):
            return self.rewrite_global_single_qubit_gate(node)
        else:
            return result.RewriteResult()

    def rewrite_qreg_new(self, node: core.QRegNew):

        addr = self.address_analysis[node.result]
        if not isinstance(addr, address.AddressReg):
            return result.RewriteResult()

        has_done_something = False
        for idx_val, qid in enumerate(addr.data):
            if qid not in self.qubit_ssa_value:
                has_done_something = True
                idx = py.constant.Constant(value=idx_val)
                qubit = core.QRegGet(node.result, idx=idx.result)
                self.qubit_ssa_value[qid] = qubit.result
                qubit.insert_after(node)
                idx.insert_after(node)

        return result.RewriteResult(has_done_something=has_done_something)

    def insert_single_qubit_noise(
        self,
        node: ir.Statement,
        qargs: ir.SSAValue,
        probs: Tuple[float, float, float, float],
    ):
        native.PauliChannel(qargs, px=probs[0], py=probs[1], pz=probs[2]).insert_before(
            node
        )
        native.AtomLossChannel(qargs, prob=probs[3]).insert_before(node)

        return result.RewriteResult(has_done_something=True)

    def rewrite_single_qubit_gate(self, node: uop.SingleQubitGate):
        probs = (
            self.gate_noise_params.local_px,
            self.gate_noise_params.local_py,
            self.gate_noise_params.local_pz,
            self.gate_noise_params.local_loss_prob,
        )
        (qargs := ilist.New(values=(node.qarg,))).insert_before(node)
        return self.insert_single_qubit_noise(node, qargs.result, probs)

    def rewrite_global_single_qubit_gate(self, node: glob.UGate):
        addrs = self.address_analysis[node.registers]
        if not isinstance(addrs, address.AddressTuple):
            return result.RewriteResult()

        qargs = []

        for addr in addrs.data:
            if not isinstance(addr, address.AddressReg):
                return result.RewriteResult()

            for qid in addr.data:
                qargs.append(self.qubit_ssa_value[qid])

        probs = (
            self.gate_noise_params.global_px,
            self.gate_noise_params.global_py,
            self.gate_noise_params.global_pz,
            self.gate_noise_params.global_loss_prob,
        )
        (qargs := ilist.New(values=tuple(qargs))).insert_before(node)
        return self.insert_single_qubit_noise(node, qargs.result, probs)

    def rewrite_parallel_single_qubit_gate(self, node: parallel.RZ | parallel.UGate):
        addrs = self.address_analysis[node.qargs]
        if not isinstance(addrs, address.AddressTuple):
            return result.RewriteResult()

        if not all(isinstance(addr, address.AddressQubit) for addr in addrs.data):
            return result.RewriteResult()

        probs = (
            self.gate_noise_params.local_px,
            self.gate_noise_params.local_py,
            self.gate_noise_params.local_pz,
            self.gate_noise_params.local_loss_prob,
        )
        assert isinstance(node.qargs, ir.ResultValue)
        assert isinstance(node.qargs.stmt, ilist.New)
        return self.insert_single_qubit_noise(node, node.qargs, probs)

    def move_noise_stmts(
        self,
        errors: Dict[Tuple[float, float, float, float], List[int]],
    ) -> list[ir.Statement]:

        nodes = []

        for probs, qubits in errors.items():
            if len(qubits) == 0:
                continue

            nodes.append(
                qargs := ilist.New(tuple(self.qubit_ssa_value[q] for q in qubits))
            )
            nodes.append(native.AtomLossChannel(qargs.result, prob=probs[3]))
            nodes.append(
                native.PauliChannel(qargs.result, px=probs[0], py=probs[1], pz=probs[2])
            )

        return nodes

    def cz_gate_noise(
        self,
        ctrls: ir.SSAValue,
        qargs: ir.SSAValue,
    ) -> list[ir.Statement]:
        return [
            native.CZPauliChannel(
                ctrls,
                qargs,
                px_ctrl=self.gate_noise_params.cz_paired_gate_px,
                py_ctrl=self.gate_noise_params.cz_paired_gate_py,
                pz_ctrl=self.gate_noise_params.cz_paired_gate_pz,
                px_qarg=self.gate_noise_params.cz_paired_gate_px,
                py_qarg=self.gate_noise_params.cz_paired_gate_py,
                pz_qarg=self.gate_noise_params.cz_paired_gate_pz,
                paired=True,
            ),
            native.CZPauliChannel(
                ctrls,
                qargs,
                px_ctrl=self.gate_noise_params.cz_unpaired_gate_px,
                py_ctrl=self.gate_noise_params.cz_unpaired_gate_py,
                pz_ctrl=self.gate_noise_params.cz_unpaired_gate_pz,
                px_qarg=self.gate_noise_params.cz_unpaired_gate_px,
                py_qarg=self.gate_noise_params.cz_unpaired_gate_py,
                pz_qarg=self.gate_noise_params.cz_unpaired_gate_pz,
                paired=False,
            ),
            native.AtomLossChannel(
                ctrls, prob=self.gate_noise_params.cz_gate_loss_prob
            ),
            native.AtomLossChannel(
                qargs, prob=self.gate_noise_params.cz_gate_loss_prob
            ),
        ]

    def rewrite_cz_gate(self, node: uop.CZ):
        qarg_addr = self.address_analysis[node.qarg]
        ctrl_addr = self.address_analysis[node.ctrl]

        if not isinstance(qarg_addr, address.AddressQubit) or not isinstance(
            ctrl_addr, address.AddressQubit
        ):
            return result.RewriteResult()

        other_qubits = sorted(
            set(self.qubit_ssa_value.keys()) - {ctrl_addr.data, qarg_addr.data}
        )
        errors = self.noise_model.parallel_cz_errors(
            [ctrl_addr.data], [qarg_addr.data], other_qubits
        )
        (ctrls := ilist.New([self.qubit_ssa_value[ctrl_addr.data]])).insert_before(node)
        (qargs := ilist.New([self.qubit_ssa_value[qarg_addr.data]])).insert_before(node)
        move_noise_nodes = self.move_noise_stmts(errors)
        gate_noise_nodes = self.cz_gate_noise(ctrls.result, qargs.result)

        for new_node in move_noise_nodes:
            new_node.insert_before(node)

        for new_node in gate_noise_nodes:
            new_node.insert_before(node)

        return result.RewriteResult(has_done_something=True)

    def rewrite_parallel_cz_gate(self, node: parallel.CZ):
        ctrls = self.address_analysis[node.ctrls]
        qargs = self.address_analysis[node.qargs]

        if not isinstance(ctrls, address.AddressTuple) or not isinstance(
            qargs, address.AddressTuple
        ):
            return result.RewriteResult()

        if not all(
            isinstance(addr, address.AddressQubit) for addr in ctrls.data
        ) or not all(isinstance(addr, address.AddressQubit) for addr in qargs.data):
            return result.RewriteResult()

        ctrl_qubits = list(map(lambda addr: addr.data, ctrls.data))
        qarg_qubits = list(map(lambda addr: addr.data, qargs.data))
        rest = sorted(set(self.qubit_ssa_value.keys()) - set(ctrl_qubits + qarg_qubits))
        errors = self.noise_model.parallel_cz_errors(ctrl_qubits, qarg_qubits, rest)

        move_noise_nodes = self.move_noise_stmts(errors)
        gate_noise_nodes = self.cz_gate_noise(node.ctrls, node.qargs)

        for new_node in move_noise_nodes:
            new_node.insert_before(node)

        for new_node in gate_noise_nodes:
            new_node.insert_before(node)

        return result.RewriteResult(has_done_something=True)



from typing import TypeVar
from dataclasses import field

from kirin import ir, interp
from bloqade.types import QubitType
from kirin.analysis import Forward, const
from kirin.analysis.forward import ForwardFrame

from .lattice import Address


class AddressAnalysis(Forward[Address]):
    """This analysis pass can be used to track the global addresses of qubits."""

    keys = ["qubit.address"]
    lattice = Address
    next_address: int = field(init=False)

    def initialize(self):
        super().initialize()
        self.next_address: int = 0
        return self

    @property
    def qubit_count(self) -> int:
        """Total number of qubits found by the analysis."""
        return self.next_address

    T = TypeVar("T")

    def get_const_value(self, typ: type[T], value: ir.SSAValue) -> T:
        if isinstance(hint := value.hints.get("const"), const.Value):
            data = hint.data
            if isinstance(data, typ):
                return hint.data
            raise interp.InterpreterError(
                f"Expected constant value <type = {typ}>, got {data}"
            )
        raise interp.InterpreterError(
            f"Expected constant value <type = {typ}>, got {value}"
        )

    def eval_stmt_fallback(
        self, frame: ForwardFrame[Address], stmt: ir.Statement
    ) -> tuple[Address, ...] | interp.SpecialValue[Address]:
        return tuple(
            (
                self.lattice.top()
                if result.type.is_subseteq(QubitType)
                else self.lattice.bottom()
            )
            for result in stmt.results
        )

    def run_method(self, method: ir.Method, args: tuple[Address, ...]):
        # NOTE: we do not support dynamic calls here, thus no need to propagate method object
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)



"""NOTE: This module is not guaranteed to be supported long-term in bloqade. We will be
moving towards a more general approach to noise modeling in the future."""

from .model import (
    GateNoiseParams as GateNoiseParams,
    TwoRowZoneModel as TwoRowZoneModel,
    MoveNoiseModelABC as MoveNoiseModelABC,
)
from .stmts import (
    PauliChannel as PauliChannel,
    CZPauliChannel as CZPauliChannel,
    AtomLossChannel as AtomLossChannel,
)
from .rewrite import RemoveNoisePass as RemoveNoisePass
from ._dialect import dialect as dialect
from ._wrappers import (
    pauli_channel as pauli_channel,
    cz_pauli_channel as cz_pauli_channel,
    atom_loss_channel as atom_loss_channel,
)



"""
Passes that deal with global gates. As of now, only one rewrite pass exists
which converts global gates to single qubit gates.
"""

from kirin import ir
from kirin.rewrite import cse, dce, walk, result
from bloqade.analysis import address
from kirin.passes.abc import Pass
from kirin.passes.fold import Fold
from bloqade.qasm2.rewrite import GlobalToUOpRule, GlobalToParallelRule
from kirin.rewrite.fixpoint import Fixpoint


class GlobalToUOP(Pass):
    """Pass to convert Global gates into single gates.

    This pass rewrites the global unitary gate from the `qasm2.glob` dialect into multiple
    single gates in the `qasm2.uop` dialect, bringing the program closer to
    conforming to standard QASM2 syntax.


    ## Usage Examples
    ```
    # Define kernel
    @qasm2.extended
    def main():
        q1 = qasm2.qreg(1)
        q2 = qasm2.qreg(2)

        theta = 1.3
        phi = 1.1
        lam = 1.2

        qasm2.glob.u(theta=theta, phi=phi, lam=lam, registers=[q1, q2])

    GlobalToUOP(dialects=main.dialects)(main)

    # Run rewrite
    GlobalToUOP(main.dialects)(main)
    ```

    The `qasm2.glob.u` statement has been rewritten to individual gates:

    ```
    qasm2.uop.u(q1[0], theta, phi, lam)
    qasm2.uop.u(q2[0], theta, phi, lam)
    qasm2.uop.u(q2[1], theta, phi, lam)
    ```
    """

    def generate_rule(self, mt: ir.Method) -> GlobalToUOpRule:
        frame, _ = address.AddressAnalysis(mt.dialects).run_analysis(mt)
        return GlobalToUOpRule(frame.entries)

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        rewriter = walk.Walk(self.generate_rule(mt))
        result = rewriter.rewrite(mt.code)

        result = walk.Walk(dce.DeadCodeElimination()).rewrite(mt.code)
        result = Fixpoint(walk.Walk(rule=cse.CommonSubexpressionElimination())).rewrite(
            mt.code
        )

        # do fold again to get proper hint for inserted const
        result = Fold(mt.dialects)(mt)
        return result


class GlobalToParallel(Pass):
    """Pass to convert Global gates into parallel gates.

    This pass rewrites the global unitary gate from the `qasm2.glob` dialect into multiple
    parallel gates in the `qasm2.parallel` dialect.


    ## Usage Examples
    ```
    # Define kernel
    @qasm2.extended
    def main():
        q1 = qasm2.qreg(1)
        q2 = qasm2.qreg(2)

        theta = 1.3
        phi = 1.1
        lam = 1.2

        qasm2.glob.u(theta=theta, phi=phi, lam=lam, registers=[q1, q2])

    GlobalToParallel(dialects=main.dialects)(main)

    # Run rewrite
    GlobalToParallel(main.dialects)(main)
    ```

    The `qasm2.glob.u` statement has been rewritten to individual gates:

    ```
    qasm2.parallel.u(theta=theta, phi=phi, lam=lam, qargs=[q1[0], q2[0], q2[1]])
    ```
    """

    def generate_rule(self, mt: ir.Method) -> GlobalToParallelRule:
        frame, _ = address.AddressAnalysis(mt.dialects).run_analysis(mt)
        return GlobalToParallelRule(frame.entries)

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        rewriter = walk.Walk(self.generate_rule(mt))
        result = rewriter.rewrite(mt.code)

        result = walk.Walk(dce.DeadCodeElimination()).rewrite(mt.code)
        result = Fixpoint(walk.Walk(rule=cse.CommonSubexpressionElimination())).rewrite(
            mt.code
        )
        # do fold again to get proper hint
        result = Fold(mt.dialects)(mt)
        return result



# %% [markdown]
# Lets do a simple example of a prototype circuit that benefits from parallelism: QAOA
# solving the MaxCut problem. For more details, see [arXiv:1411.4028](https://arxiv.org/abs/1411.4028)
# and the considerable literature that has developed around this algorithm.

import math
from typing import Any

import kirin
import networkx as nx
from bloqade import qasm2
from kirin.dialects import py, ilist

pi = math.pi

# %% [markdown]
# MaxCut is a combinatorial graph problem that seeks to bi-partition the nodes of some
# graph G such that the number of edges between the two partitions is maximized.
# Here, we choose a random 3 regular graph with 32 nodes [ref](https://journals.aps.org/pra/abstract/10.1103/PhysRevA.103.042612)
N = 32
G = nx.random_regular_graph(3, N, seed=42)


# %% [markdown]
# To build the quantum program, we use a builder function and use closure to pass variables
# inside of the kernel function (kirin methods).
# In this case, the two variables that are passed inside are the edges and nodes of the graph.
#
# The QAOA first prepares the |+> state as a superposition of all possible bitstrings,
# then repeats between the (diagonal) cost function and the mixer X with angles gamma and beta.
# It is parameterized by gamma and betas, which are each the p length lists of angles.
#
# Lets first implement the sequential version of the QAOA algorithm, which
# does not inform any parallelism to the compiler.


def qaoa_sequential(G: nx.Graph) -> kirin.ir.Method:

    edges = list(G.edges)
    nodes = list(G.nodes)
    N = len(nodes)

    @qasm2.extended
    def kernel(gamma: ilist.IList[float, Any], beta: ilist.IList[float, Any]):
        # Initialize the register in the |+> state
        qreg = qasm2.qreg(N)
        for i in range(N):  # structural control flow is native to the Kirin compiler
            qasm2.h(qreg[i])

        # Repeat the cost and mixer layers
        for i in range(len(gamma)):
            # The cost layer, which corresponds to a ZZ(phase) gate applied
            # to each edge of the graph
            for j in range(len(edges)):
                edge = edges[j]
                qasm2.cx(qreg[edge[0]], qreg[edge[1]])
                qasm2.rz(qreg[edge[1]], gamma[i])
                qasm2.cx(qreg[edge[0]], qreg[edge[1]])
            # The mixer layer, which corresponds to a X(phase) gate applied
            # to each node of the graph
            for j in range(N):
                qasm2.rx(qreg[j], beta[i])

        return qreg

    return kernel


# %% [markdown]
# Next, lets implement a SIMD (Single Instruction, Multiple Data) version of the QAOA algorithm,
# which effectively represents the parallelism in the QAOA algorithm.
# This can be done by coloring the (commuting) ZZ(phase) gates into groups with non-overlapping
# sets of qubits, and then applying each of those groups in parallel.
# By [Vizing's theorem](https://en.wikipedia.org/wiki/Vizing%27s_theorem) the edges of a graph
# can efficiently be colored into $\Delta+1$ colors, where $\Delta$ is the maximum degree of the graph.
# Unfortunately, networkx does not have a native implementation of the algorithm so instead we use
# the lesser [Brooks' theorem]https://en.wikipedia.org/wiki/Brooks%27_theorem) to color the edges
# using an equitable coloring of the line graph.


def qaoa_simd(G: nx.Graph) -> kirin.ir.Method:

    nodes = list(G.nodes)

    # Note that graph computation is happening /outside/ the kernel function:
    # this is a computation that occurs on your laptop in Python when you generate
    # a program, as opposed to on a piece of quantum hardware, which is what
    # occurs inside of the kernel.
    Gline = nx.line_graph(G)
    colors = nx.algorithms.coloring.equitable_color(Gline, num_colors=5)
    left_ids = ilist.IList(
        [
            ilist.IList([edge[0] for edge in G.edges if colors[edge] == i])
            for i in range(5)
        ]
    )
    right_ids = ilist.IList(
        [
            ilist.IList([edge[1] for edge in G.edges if colors[edge] == i])
            for i in range(5)
        ]
    )
    # We can use composition of kernel functions to simplify repeated code.
    # Small snippets (say, the CX gate) can be written once and then called
    # many times.

    @qasm2.extended
    def parallel_h(qargs: ilist.IList[qasm2.Qubit, Any]):
        qasm2.parallel.u(qargs=qargs, theta=pi / 2, phi=0.0, lam=pi)

    # A parallel CX gate is equivalently a parallel H gate, followed by a parallel CZ gate,
    # followed by another parallel H. the CZ can be done in any order as they permute.
    @qasm2.extended
    def parallel_cx(
        ctrls: ilist.IList[qasm2.Qubit, Any], qargs: ilist.IList[qasm2.Qubit, Any]
    ):
        parallel_h(qargs)
        qasm2.parallel.cz(ctrls, qargs)
        parallel_h(qargs)

    @qasm2.extended
    def parallel_cz_phase(
        ctrls: ilist.IList[qasm2.Qubit, Any],
        qargs: ilist.IList[qasm2.Qubit, Any],
        gamma: float,
    ):
        parallel_cx(ctrls, qargs)
        qasm2.parallel.rz(qargs, gamma)
        parallel_cx(ctrls, qargs)

    @qasm2.extended
    def kernel(gamma: ilist.IList[float, Any], beta: ilist.IList[float, Any]):
        # Declare the register and set it to the |+> state
        qreg = qasm2.qreg(len(nodes))
        # qasm2.glob.u(theta=pi / 2, phi=0.0, lam=pi,registers=[qreg])

        def get_qubit(x: int):
            return qreg[x]

        all_qubits = ilist.map(fn=get_qubit, collection=range(N))

        parallel_h(all_qubits)

        for i in range(len(gamma)):  # For each QAOA layer...
            # Do the ZZ phase gates...
            for cind in range(
                5
            ):  # by applying a parallel CZ phase gate in parallel for each color,
                ctrls = ilist.map(fn=get_qubit, collection=left_ids[cind])
                qargs = ilist.map(fn=get_qubit, collection=right_ids[cind])
                parallel_cz_phase(ctrls, qargs, gamma[i])
            # ...then, do an X phase gate. Observe that because this happens on every
            # qubit, we can do a global rotation, which is higher fidelity than
            # parallel local rotations.
            # qasm2.glob.u(theta=beta[i],phi=0.0,lam=0.0,registers=[qreg])
            qasm2.parallel.u(qargs=all_qubits, theta=beta[i], phi=0.0, lam=0.0)

        return qreg

    return kernel


# %%
print("--- Sequential ---")
qaoa_sequential(G).code.print()

# %%
kernel = qaoa_simd(G)

print("\n\n--- Simd ---")
kernel.print()


# %%
@qasm2.extended
def main():
    kernel([0.1, 0.2], [0.3, 0.4])


# %%
target = qasm2.emit.QASM2(
    main_target=qasm2.main.union(
        [qasm2.dialects.parallel, qasm2.dialects.glob, ilist, py.constant]
    )
)
ast = target.emit(main)
qasm2.parse.pprint(ast)



from bloqade import qasm2


def test_global_allow_global():

    @qasm2.extended
    def glob_u():
        qreg = qasm2.qreg(3)
        qreg1 = qasm2.qreg(3)
        qasm2.glob.u(theta=0.1, phi=0.2, lam=0.3, registers=[qreg, qreg1])

    target = qasm2.emit.QASM2(
        allow_global=True,
        allow_parallel=False,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(glob_u)
    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.glob,qasm2.indexing,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
qreg qreg1[3];
glob.U(0.1, 0.2, 0.3) {qreg, qreg1}
"""
    )


def test_global_allow_global_allow_para():

    @qasm2.extended
    def glob_u():
        qreg = qasm2.qreg(3)
        qreg1 = qasm2.qreg(3)
        qasm2.glob.u(theta=0.1, phi=0.2, lam=0.3, registers=[qreg, qreg1])

    target = qasm2.emit.QASM2(
        allow_global=True,
        allow_parallel=True,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(glob_u)
    print(qasm2_str)
    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.glob,qasm2.indexing,qasm2.parallel,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
qreg qreg1[3];
glob.U(0.1, 0.2, 0.3) {qreg, qreg1}
"""
    )


def test_global():

    @qasm2.extended
    def glob_u():
        qreg = qasm2.qreg(3)
        qreg1 = qasm2.qreg(3)
        qasm2.glob.u(theta=0.1, phi=0.2, lam=0.3, registers=[qreg, qreg1])

    target = qasm2.emit.QASM2(
        allow_global=False,
        allow_parallel=False,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(glob_u)
    assert (
        qasm2_str
        == """OPENQASM 2.0;
include "qelib1.inc";
qreg qreg[3];
qreg qreg1[3];
U(0.1, 0.2, 0.3) qreg1[2];
U(0.1, 0.2, 0.3) qreg1[1];
U(0.1, 0.2, 0.3) qreg1[0];
U(0.1, 0.2, 0.3) qreg[2];
U(0.1, 0.2, 0.3) qreg[1];
U(0.1, 0.2, 0.3) qreg[0];
"""
    )


def test_global_allow_para():

    @qasm2.extended
    def glob_u():
        qreg = qasm2.qreg(3)
        qreg1 = qasm2.qreg(3)
        qasm2.glob.u(theta=0.1, phi=0.2, lam=0.3, registers=[qreg, qreg1])

    target = qasm2.emit.QASM2(
        allow_global=False,
        allow_parallel=True,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(glob_u)

    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.indexing,qasm2.parallel,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
qreg qreg1[3];
parallel.U(0.1, 0.2, 0.3) {
  qreg[0];
  qreg[1];
  qreg[2];
  qreg1[0];
  qreg1[1];
  qreg1[2];
}
"""
    )


def test_para():

    @qasm2.extended
    def para_u():
        qreg = qasm2.qreg(3)
        qasm2.parallel.u(theta=0.1, phi=0.2, lam=0.3, qargs=[qreg[0], qreg[1]])

    para_u.print()

    target = qasm2.emit.QASM2(
        allow_parallel=False,
        allow_global=False,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(para_u)
    assert (
        qasm2_str
        == """OPENQASM 2.0;
include "qelib1.inc";
qreg qreg[3];
U(0.1, 0.2, 0.3) qreg[1];
U(0.1, 0.2, 0.3) qreg[0];
"""
    )


def test_para_allow_para():

    @qasm2.extended
    def para_u():
        qreg = qasm2.qreg(3)
        qasm2.parallel.u(theta=0.1, phi=0.2, lam=0.3, qargs=[qreg[0], qreg[1]])

    para_u.print()

    target = qasm2.emit.QASM2(
        allow_parallel=True,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(para_u)
    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.indexing,qasm2.parallel,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
parallel.U(0.1, 0.2, 0.3) {
  qreg[0];
  qreg[1];
}
"""
    )


def test_para_allow_para_allow_global():

    @qasm2.extended
    def para_u():
        qreg = qasm2.qreg(3)
        qasm2.parallel.u(theta=0.1, phi=0.2, lam=0.3, qargs=[qreg[0], qreg[1]])

    para_u.print()

    target = qasm2.emit.QASM2(
        allow_parallel=True,
        allow_global=True,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(para_u)
    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.glob,qasm2.indexing,qasm2.parallel,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
parallel.U(0.1, 0.2, 0.3) {
  qreg[0];
  qreg[1];
}
"""
    )


def test_para_allow_global():

    @qasm2.extended
    def para_u():
        qreg = qasm2.qreg(3)
        qasm2.parallel.u(theta=0.1, phi=0.2, lam=0.3, qargs=[qreg[0], qreg[1]])

    para_u.print()

    target = qasm2.emit.QASM2(
        allow_parallel=False,
        allow_global=True,
        custom_gate=True,
    )
    qasm2_str = target.emit_str(para_u)
    print(qasm2_str)
    assert (
        qasm2_str
        == """KIRIN {func,lowering.call,lowering.func,py.ilist,qasm2.core,qasm2.expr,qasm2.glob,qasm2.indexing,qasm2.uop,scf};
include "qelib1.inc";
qreg qreg[3];
U(0.1, 0.2, 0.3) qreg[1];
U(0.1, 0.2, 0.3) qreg[0];
"""
    )



"""Rewrite qasm dialects into py dialects."""

from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import Walk, Fixpoint
from kirin.dialects import py, math
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult
from bloqade.qasm2.dialects import core, expr


class _QASM2Py(RewriteRule):
    """Rewrite qasm dialects into py dialects."""

    UNARY_OPS = {
        expr.Neg: py.USub,
        expr.Sin: math.sin,
        expr.Cos: math.cos,
        expr.Tan: math.tan,
        expr.Exp: math.exp,
        expr.Sqrt: math.sqrt,
    }

    BINARY_OPS = {
        expr.Add: py.Add,
        expr.Sub: py.Sub,
        expr.Mul: py.Mult,
        expr.Div: py.Div,
        expr.Pow: py.Pow,
    }

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, (expr.ConstInt, expr.ConstFloat)):
            node.replace_by(py.Constant(value=node.value))
            return RewriteResult(has_done_something=True)
        elif isinstance(node, expr.Neg):
            node.replace_by(self.UNARY_OPS[type(node)](value=node.value))
            return RewriteResult(has_done_something=True)
        elif isinstance(node, (expr.Sin, expr.Cos, expr.Tan, expr.Exp, expr.Sqrt)):
            node.replace_by(self.UNARY_OPS[type(node)](x=node.value))
            return RewriteResult(has_done_something=True)
        elif isinstance(node, (expr.Add, expr.Sub, expr.Mul, expr.Div, expr.Pow)):
            node.replace_by(self.BINARY_OPS[type(node)](lhs=node.lhs, rhs=node.rhs))
            return RewriteResult(has_done_something=True)
        elif isinstance(node, core.CRegEq):
            node.replace_by(py.cmp.Eq(node.lhs, node.rhs))
            return RewriteResult(has_done_something=True)
        else:
            return RewriteResult()


class QASM2Py(Pass):

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        return Fixpoint(Walk(_QASM2Py())).rewrite(mt.code)



"""Generate the code reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files

SRC_PATH = "src"

skip_keywords = [
    "julia",  ## [KHW] skip for now since we didn't have julia codegen rdy
    "builder/base",  ## hiding from user
    "builder/terminate",  ## hiding from user
    "ir/tree_print",  ## hiding from user
    "ir/visitor",  ## hiding from user
    "codegen/",  ## hiding from user
    "builder/factory",  ## hiding from user
    "builder_old",  ## deprecated from user
    "task_old",  ## deprecated from user
    "visualization",  ## hiding from user
    "submission/capabilities",  ## hiding from user
    "submission/quera_api_client",
]

nav = mkdocs_gen_files.Nav()
for path in sorted(Path(SRC_PATH).rglob("*.py")):
    module_path = path.relative_to(SRC_PATH).with_suffix("")
    doc_path = path.relative_to(SRC_PATH).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    iskip = False

    for kwrd in skip_keywords:
        if kwrd in str(doc_path):
            iskip = True
            break
    if iskip:
        print("[Ignore]", str(doc_path))
        continue

    print("[>]", str(doc_path))

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1].startswith("_"):
        continue

    nav[parts] = doc_path.as_posix()
    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, ".." / path)

with mkdocs_gen_files.open("reference/SUMMARY.txt", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.dialects import ilist
from bloqade.qasm2.parse import ast
from bloqade.qasm2.types import QRegType
from bloqade.qasm2.emit.gate import EmitQASM2Gate, EmitQASM2Frame
from bloqade.analysis.schedule import DagScheduleAnalysis

dialect = ir.Dialect("qasm2.glob")


@statement(dialect=dialect)
class UGate(ir.Statement):
    name = "ugate"
    traits = frozenset({ir.FromPythonCall()})
    registers: ir.SSAValue = info.argument(ilist.IListType[QRegType])
    theta: ir.SSAValue = info.argument(types.Float)
    phi: ir.SSAValue = info.argument(types.Float)
    lam: ir.SSAValue = info.argument(types.Float)


@dialect.register(key="qasm2.schedule.dag")
class Glob(interp.MethodTable):
    @interp.impl(UGate)
    def ugate(self, interp: DagScheduleAnalysis, frame: interp.Frame, stmt: UGate):
        interp.update_dag(stmt, [stmt.registers])
        return ()


@dialect.register(key="emit.qasm2.gate")
class GlobEmit(interp.MethodTable):
    @interp.impl(UGate)
    def ugate(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: UGate):
        registers = [
            emit.assert_node(ast.Name, reg) for reg in frame.get(stmt.registers)
        ]
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        frame.body.append(
            ast.GlobUGate(theta=theta, phi=phi, lam=lam, registers=registers)
        )
        return ()



import io

from kirin import ir
from rich.console import Console
from kirin.analysis import CallGraph
from kirin.dialects import ilist
from bloqade.qasm2.parse import ast, pprint
from bloqade.qasm2.passes.fold import QASM2Fold
from bloqade.qasm2.passes.glob import GlobalToParallel
from bloqade.qasm2.passes.py2qasm import Py2QASM
from bloqade.qasm2.passes.parallel import ParallelToUOp

from .gate import EmitQASM2Gate
from .main import EmitQASM2Main


class QASM2:
    """QASM2 target for Bloqade kernels.

    QASM2 target that accepts a Bloqade kernel and produces an AST that you can then obtain a string for printing or saving as a file.
    """

    def __init__(
        self,
        qelib1: bool = True,
        allow_parallel: bool = False,
        allow_global: bool = False,
        custom_gate: bool = True,
    ) -> None:
        """Initialize the QASM2 target.

        Args:
            allow_parallel (bool):
                Allow parallel gate in the resulting QASM2 AST. Defaults to `False`.
                In the case its False, and the input kernel uses parallel gates, they will get rewrite into uop gates.

            allow_global (bool):
                Allow global gate in the resulting QASM2 AST. Defaults to `False`.
                In the case its False, and the input kernel uses global gates, they will get rewrite into parallel gates.
                If both `allow_parallel` and `allow_global` are False, the input kernel will be rewritten to use uop gates.

            qelib1 (bool):
                Include the `include "qelib1.inc"` line in the resulting QASM2 AST that's
                submitted to qBraid. Defaults to `True`.
            custom_gate (bool):
                Include the custom gate definitions in the resulting QASM2 AST. Defaults to `True`. If `False`, all the qasm2.gate will be inlined.



        """
        from bloqade import qasm2

        self.main_target = qasm2.main
        self.gate_target = qasm2.gate

        self.qelib1 = qelib1
        self.custom_gate = custom_gate
        self.allow_parallel = allow_parallel
        self.allow_global = allow_global

        if allow_parallel:
            self.main_target = self.main_target.add(qasm2.dialects.parallel)
            self.gate_target = self.gate_target.add(qasm2.dialects.parallel)

        if allow_global:
            self.main_target = self.main_target.add(qasm2.dialects.glob)
            self.gate_target = self.gate_target.add(qasm2.dialects.glob)

        if allow_global or allow_parallel:
            self.main_target = self.main_target.add(ilist)
            self.gate_target = self.gate_target.add(ilist)

    def emit(self, entry: ir.Method) -> ast.MainProgram:
        """Emit a QASM2 AST from the Bloqade kernel.

        Args:
            entry (ir.Method):
                The Bloqade kernel to convert to the QASM2 AST

        Returns:
            ast.MainProgram:
                A QASM2 AST object

        """
        assert len(entry.args) == 0, "entry method should not have arguments"

        # make a cloned instance of kernel
        entry = entry.similar()
        QASM2Fold(entry.dialects, inline_gate_subroutine=not self.custom_gate).fixpoint(
            entry
        )

        if not self.allow_global:
            # rewrite global to parallel
            GlobalToParallel(dialects=entry.dialects)(entry)

        if not self.allow_parallel:
            # rewrite parallel to uop
            ParallelToUOp(dialects=entry.dialects)(entry)

        Py2QASM(entry.dialects)(entry)
        target_main = EmitQASM2Main(self.main_target)
        target_main.run(
            entry, tuple(ast.Name(name) for name in entry.arg_names[1:])
        ).expect()

        main_program = target_main.output
        assert main_program is not None, f"failed to emit {entry.sym_name}"

        extra = []
        if self.qelib1:
            extra.append(ast.Include("qelib1.inc"))

        if self.custom_gate:
            cg = CallGraph(entry)
            target_gate = EmitQASM2Gate(self.gate_target)

            for _, fn in cg.defs.items():
                if fn is entry:
                    continue

                fn = fn.similar()
                QASM2Fold(fn.dialects).fixpoint(fn)

                if not self.allow_global:
                    # rewrite global to parallel
                    GlobalToParallel(dialects=fn.dialects)(fn)

                if not self.allow_parallel:
                    # rewrite parallel to uop
                    ParallelToUOp(dialects=fn.dialects)(fn)

                Py2QASM(fn.dialects)(fn)

                target_gate.run(
                    fn, tuple(ast.Name(name) for name in fn.arg_names[1:])
                ).expect()
                assert target_gate.output is not None, f"failed to emit {fn.sym_name}"
                extra.append(target_gate.output)

        main_program.statements = extra + main_program.statements
        return main_program

    def emit_str(self, entry: ir.Method) -> str:
        """Emit a QASM2 AST from the Bloqade kernel.

        Args:
            entry (ir.Method):
                The Bloqade kernel to convert to the QASM2 AST

        Returns:
            str:
                A string with the QASM2 representation of the kernel

        """
        console = Console(
            file=io.StringIO(),
            force_terminal=False,
            force_interactive=False,
            force_jupyter=False,
            record=True,
        )
        pprint(self.emit(entry), console=console)
        return console.export_text()



from bloqade import qasm2
from bloqade.noise import native
from bloqade.noise.native.rewrite import RemoveNoisePass

simulation = qasm2.extended.add(native)


def test():

    @simulation
    def test_atom_loss():
        q = qasm2.qreg(2)
        native.atom_loss_channel([q[0], q[1]], prob=0.7)
        native.cz_pauli_channel(
            [q[0]],
            [q[1]],
            paired=False,
            px_ctrl=0.1,
            py_ctrl=0.4,
            pz_ctrl=0.3,
            px_qarg=0.2,
            py_qarg=0.2,
            pz_qarg=0.2,
        )  # no noise here
        qasm2.parallel.cz([q[0]], [q[1]])
        native.atom_loss_channel([q[0], q[1]], prob=0.7)
        native.cz_pauli_channel(
            [q[0]],
            [q[1]],
            paired=False,
            px_ctrl=0.1,
            py_ctrl=0.4,
            pz_ctrl=0.3,
            px_qarg=0.2,
            py_qarg=0.2,
            pz_qarg=0.2,
        )
        qasm2.parallel.cz([q[0]], [q[1]])
        return q

    @simulation
    def tests():
        q = qasm2.qreg(2)
        qasm2.parallel.cz([q[0]], [q[1]])
        qasm2.parallel.cz([q[0]], [q[1]])
        return q

    RemoveNoisePass(simulation)(test_atom_loss)
    assert test_atom_loss.callable_region.is_structurally_equal(tests.callable_region)



import abc
import math
from typing import Dict, List, Tuple
from dataclasses import field, dataclass
from collections.abc import Sequence


@dataclass(frozen=True)
class GateNoiseParams:
    """Parameters for gate noise."""

    local_px: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-X error during a local single qubit gate operation."""
    local_py: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Y error during a local single qubit gate operation."""
    local_pz: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Z error during a local single qubit gate operation."""
    local_loss_prob: float = field(default=1e-4, kw_only=True)
    """The error probability for a loss during a local single qubit gate operation."""

    global_px: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-X error during a global single qubit gate operation."""
    global_py: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Y error during a global single qubit gate operation."""
    global_pz: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Z error during a global single qubit gate operation."""
    global_loss_prob: float = field(default=1e-3, kw_only=True)
    """The error probability for a loss during a global single qubit gate operation."""

    cz_paired_gate_px: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-X error during CZ gate operation when two qubits are within blockade radius."""
    cz_paired_gate_py: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Y error during CZ gate operation when two qubits are within blockade radius."""
    cz_paired_gate_pz: float = field(default=1e-3, kw_only=True)
    """The error probability for a Pauli-Z error during CZ gate operation when two qubits are within blockade radius."""
    cz_gate_loss_prob: float = field(default=1e-3, kw_only=True)
    """The error probability for a loss during CZ gate operation when two qubits are within blockade radius."""

    cz_unpaired_gate_px: float = field(default=1e-3, kw_only=True)
    """The error probability for Pauli-X error during CZ gate operation when another qubit is not within blockade radius."""
    cz_unpaired_gate_py: float = field(default=1e-3, kw_only=True)
    """The error probability for Pauli-Y error during CZ gate operation when another qubit is not within blockade radius."""
    cz_unpaired_gate_pz: float = field(default=1e-3, kw_only=True)
    """The error probability for Pauli-Z error during CZ gate operation when another qubit is not within blockade radius."""
    cz_unpaired_loss_prob: float = field(default=1e-3, kw_only=True)
    """The error probability for a loss during CZ gate operation when another qubit is not within blockade radius."""


@dataclass(frozen=True)
class MoveNoiseParams:
    move_px_rate: float = field(default=1e-6, kw_only=True)
    """The error rate (prob/microsecond) for a Pauli-X error during a move operation."""
    move_py_rate: float = field(default=1e-6, kw_only=True)
    """The error rate e (prob/microsecond) for a Pauli-Y error during a move operation."""
    move_pz_rate: float = field(default=1e-6, kw_only=True)
    """The error rate e (prob/microsecond) for a Pauli-Z error during a move operation."""
    move_loss_rate: float = field(default=1e-6, kw_only=True)
    """The error rate e (prob/microsecond) for a loss during a move operation."""

    pick_px: float = field(default=1e-3, kw_only=True)
    """The error rate (prob per pick operation)  for a Pauli-X error during a pick operation."""
    pick_py: float = field(default=1e-3, kw_only=True)
    """The error rate (prob per pick operation) for a Pauli-Y error during a pick operation."""
    pick_pz: float = field(default=1e-3, kw_only=True)
    """The error rate (prob per pick operation) for a Pauli-Z error during a pick operation."""
    pick_loss_prob: float = field(default=1e-4, kw_only=True)
    """The error rate for a loss during a pick operation."""

    move_speed: float = field(default=5e-1, kw_only=True)
    """Maximum speed of the qubits during a move operation."""
    storage_spacing: float = field(default=4.0, kw_only=True)
    """Spacing between the qubits in the storage zone."""


@dataclass
class MoveNoiseModelABC(abc.ABC):
    """Abstract base class for noise based on atom movement.

    This class defines the interface for a noise model. The gate noise is calculated form the parameters
    provided in this dataclass which can be updated when inheriting from this class. The move error is
    calculated by implementing the parallel_cz_errors method which takes a set of ctrl and qarg qubits
    and returns a noise model for all the qubits. The noise model is a dictionary with the keys being the
    error rates for the qubits and the values being the list of qubits that the error rate applies to.

    Once implemented the class can be used with the NoisePass to analyze a circuit and apply the noise
    model to the circuit.

    NOTE: This model is not guaranteed to be supported long-term in bloqade. We will be
    moving towards a more general approach to noise modeling in the future.

    """

    params: MoveNoiseParams = field(default_factory=MoveNoiseParams)
    """Parameters for calculating move noise."""

    @classmethod
    @abc.abstractmethod
    def parallel_cz_errors(
        cls, ctrls: List[int], qargs: List[int], rest: List[int]
    ) -> Dict[Tuple[float, float, float, float], List[int]]:
        """Takes a set of ctrls and qargs and returns a noise model for all qubits."""
        pass

    @staticmethod
    def poisson_pauli_prob(rate: float, duration: float) -> float:
        """Calculate the number of noise events and their probabilities for a given rate and duration."""
        assert duration >= 0, "Duration must be non-negative"
        assert rate >= 0, "Rate must be non-negative"
        return 0.5 * (1 - math.exp(-2 * rate * duration))

    @classmethod
    def join_binary_probs(cls, p1: float, *args: float) -> float:
        """Merge the probabilities of an event happening if the event can only happen once.

        For example, finding the effective probability of losing an atom from multiple sources, since
        a qubit can only happen once. This is done by using the formula:

        p = p1 * (1 - p2) + p2 * (1 - p1)

        applied recursively to all the probabilities in the list.

        Args:
            p1 (float): The probability of the event happening.
            arg (float): The probabilities of the event happening from other sources.

        Returns:
            float: The effective probability of the event happening.

        """
        if len(args) == 0:
            return p1
        else:
            p2 = cls.join_binary_probs(*args)
            return p1 * (1 - p2) + p2 * (1 - p1)


@dataclass
class TwoRowZoneModel(MoveNoiseModelABC):
    """This model assumes that the qubits are arranged in a single storage row with a row corresponding to a gate zone below it.

    The CZ gate noise is calculated using the following heuristic: The idle error is calculated by the total duration require
    to do the move and entable the qubits. Not every pair can be entangled at the same time, so we first deconflict the qargs
    assuming by finding subsets in which both the ctrl and the qarg qubits are in ascending order. This breaks the pairs into
    groups that can be moved and entangled separately. We then take each group and assign each pair to a gate zone slot. The
    slots are allocated by starting from the middle of the atoms and moving outwards making sure to keep the ctrl qubits in
    ascending order. The time to move a group is calculated by finding the maximum travel distance of the qarg and ctrl qubits
    and dviding by the move speed. The total move time is the sum of all the group move times. The error rate for all the qubits
    is then calculated by using the poisson_pauli_prob function. An additional error for the pick operation is calculated by
    joining the binary probabilities of the pick operation and the move operation.

    """

    gate_zone_y_offset: float = 20.0
    gate_spacing: float = 20.0

    def deconflict(
        self, ctrls: List[int], qargs: List[int]
    ) -> List[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
        """Return a list of groups of ctrl and qarg qubits that can be moved and entangled separately."""
        # sort by ctrl qubit first to guarantee that they will be in ascending order
        sorted_pairs = sorted(zip(ctrls, qargs))

        groups = []
        # group by qarg only putting it in a group if the qarg is greater than the last qarg in the group
        # thus ensuring that the qargs are in ascending order
        while len(sorted_pairs) > 0:
            ctrl, qarg = sorted_pairs.pop(0)

            found = False
            for group in groups:
                if group[-1][1] < qarg:
                    group.append((ctrl, qarg))
                    found = True
                    break
            if not found:
                groups.append([(ctrl, qarg)])

        return [tuple(zip(*group)) for group in groups]

    def assign_gate_slots(
        self, ctrls: Sequence[int], qargs: Sequence[int]
    ) -> Dict[int, Tuple[int, int]]:
        """Allocate slots for the qubits to move to. start from middle of atoms and move outwards
        making sure to keep the ctrl qubits in ascending order.

        Note that we can do this because the move strategy is to move the ctrl qubits separately
        from the qarg qubits, thus we don't have to worry about qarg qubits crossing the ctrl qubits
        and vice versa. We pick the median of all the atoms because it distributes the qubits
        as evenly as possible over the gate zone.

        """
        assert len(ctrls) == len(qargs), "Number of ctrls and qargs must be equal"
        addr_pairs = sorted(zip(ctrls, qargs), key=lambda x: x[0])
        # sort by the distance between the ctrl and qarg qubits

        ctrls, qargs = list(zip(*addr_pairs))

        n_ctrls = len(ctrls)

        ctrl_median = (
            ctrls[n_ctrls // 2]
            if n_ctrls % 2 == 1
            else (ctrls[n_ctrls // 2 - 1] + ctrls[n_ctrls // 2]) / 2
        )

        all_addr = sorted(ctrls + qargs)
        spatial_median = self.params.storage_spacing * (all_addr[0] + all_addr[-1]) / 2

        addr_pairs.sort(key=lambda x: abs(x[0] - ctrl_median))

        slots = {}
        med_slot = round(spatial_median / self.gate_spacing)

        left_slot = med_slot
        right_slot = med_slot
        slots[med_slot] = addr_pairs.pop(0)
        while addr_pairs:
            ctrl, qarg = addr_pairs.pop(0)

            if ctrl < ctrl_median:
                slots[left_slot := left_slot - 1] = (ctrl, qarg)
            else:
                slots[right_slot := right_slot + 1] = (ctrl, qarg)

        return slots

    def calculate_move_duration(self, slots: Dict[int, Tuple[int, int]]) -> float:
        """Calculate the time it takes to move the qubits from the ctrl to the qarg qubits."""

        qarg_x_distance = float("-inf")
        ctrl_x_distance = float("-inf")

        for slot, (ctrl, qarg) in slots.items():
            qarg_x_distance = max(
                qarg_x_distance,
                abs(qarg * self.params.storage_spacing - slot * self.gate_spacing),
            )
            ctrl_x_distance = max(
                ctrl_x_distance,
                abs(ctrl * self.params.storage_spacing - slot * self.gate_spacing),
            )

        qarg_max_distance = math.sqrt(qarg_x_distance**2 + self.gate_zone_y_offset**2)
        ctrl_max_distance = math.sqrt(
            ctrl_x_distance**2 + (self.gate_zone_y_offset - 3) ** 2
        )

        return (qarg_max_distance + ctrl_max_distance) / self.params.move_speed

    def parallel_cz_errors(
        self, ctrls: List[int], qargs: List[int], rest: List[int]
    ) -> Dict[Tuple[float, float, float, float], List[int]]:
        """Apply parallel gates by moving ctrl qubits to qarg qubits."""
        groups = self.deconflict(ctrls, qargs)
        slots = [self.assign_gate_slots(*group) for group in groups]

        move_duration = sum(map(self.calculate_move_duration, slots))

        px_time = self.poisson_pauli_prob(self.params.move_px_rate, move_duration)
        py_time = self.poisson_pauli_prob(self.params.move_py_rate, move_duration)
        px_time = self.poisson_pauli_prob(self.params.move_pz_rate, move_duration)
        p_loss_time = self.poisson_pauli_prob(self.params.move_loss_rate, move_duration)

        errors = {(px_time, py_time, px_time, p_loss_time): rest}

        px_moved = self.join_binary_probs(self.params.pick_px, px_time)
        py_moved = self.join_binary_probs(self.params.pick_py, py_time)
        pz_moved = self.join_binary_probs(self.params.pick_pz, px_time)
        p_loss_moved = self.join_binary_probs(self.params.pick_loss_prob, p_loss_time)

        errors[(px_moved, py_moved, pz_moved, p_loss_moved)] = sorted(ctrls + qargs)

        return errors



from typing import Dict, List, Tuple, Sequence
from dataclasses import field, dataclass

from kirin import ir, types, passes
from bloqade import noise, qasm2
from bloqade.qbraid import schema
from kirin.dialects import func, ilist
from bloqade.qasm2.dialects import glob, parallel


@ir.dialect_group(
    [func, qasm2.core, qasm2.uop, parallel, glob, qasm2.expr, noise.native, ilist]
)
def qbraid_noise(
    self,
):
    fold_pass = passes.Fold(self)
    typeinfer_pass = passes.TypeInfer(self)

    def run_pass(
        method: ir.Method,
        *,
        fold: bool = True,
    ):
        method.verify()

        if fold:
            fold_pass(method)

        typeinfer_pass(method)
        method.code.typecheck()

    return run_pass


@dataclass
class Lowering:
    qubit_list: List[ir.SSAValue] = field(init=False, default_factory=list)
    qubit_id_map: Dict[int, ir.SSAValue] = field(init=False, default_factory=dict)
    bit_id_map: Dict[int, ir.SSAValue] = field(init=False, default_factory=dict)
    block_list: List[ir.Statement] = field(init=False, default_factory=list)

    def lower(
        self,
        sym_name: str,
        noise_model: schema.NoiseModel,
        return_qreg: bool = False,
    ) -> ir.Method:
        """Lower the noise model to a method.

        Args:
            name (str): The name of the method to generate.
            return_qreg (bool): Use the quantum register as the return value.

        Returns:
            Method: The generated kirin method.

        """
        self.process_noise_model(noise_model, return_qreg)
        block = ir.Block(stmts=self.block_list)
        ret_type = qasm2.types.QRegType if return_qreg else qasm2.types.CRegType
        block.args.append_from(types.MethodType[[], ret_type], name=f"{sym_name}_self")
        region = ir.Region(block)
        func_stmt = func.Function(
            sym_name=sym_name,
            signature=func.Signature(inputs=(), output=qasm2.types.QRegType),
            body=region,
        )

        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=sym_name,
            dialects=qbraid_noise,
            code=func_stmt,
            arg_names=[],
        )
        qbraid_noise.run_pass(mt)  # type: ignore
        return mt

    def process_noise_model(self, noise_model: schema.NoiseModel, return_qreg: bool):
        num_qubits = self.lower_number(noise_model.num_qubits)

        reg = qasm2.core.QRegNew(num_qubits)
        creg = qasm2.core.CRegNew(num_qubits)
        self.block_list.append(reg)
        self.block_list.append(creg)

        for idx_value, qubit in enumerate(noise_model.all_qubits):
            idx = self.lower_number(idx_value)
            self.block_list.append(qubit_stmt := qasm2.core.QRegGet(reg.result, idx))
            self.block_list.append(bit_stmt := qasm2.core.CRegGet(creg.result, idx))

            self.qubit_id_map[qubit] = qubit_stmt.result
            self.bit_id_map[qubit] = bit_stmt.result
            self.qubit_list.append(qubit_stmt.result)

        for gate_event in noise_model.gate_events:
            self.process_gate_event(gate_event)

        if return_qreg:
            self.block_list.append(func.Return(reg.result))
        else:
            self.block_list.append(func.Return(creg.result))

    def process_gate_event(self, node: schema.GateEvent):
        self.lower_atom_loss(node.error.survival_prob)

        if isinstance(node.operation, schema.CZ):
            assert isinstance(node.error, schema.CZError), "Only CZError is supported"
            self.process_cz_pauli_error(node.operation.participants, node.error)
            self.lower_cz_gates(node.operation)
        else:

            error = node.error
            assert isinstance(
                error, schema.SingleQubitError
            ), "Only SingleQubitError is supported"
            self.lower_pauli_errors(error.operator_error)

            operation = node.operation
            assert isinstance(
                operation,
                (
                    schema.GlobalW,
                    schema.LocalW,
                    schema.GlobalRz,
                    schema.LocalRz,
                    schema.Measurement,
                ),
            ), f"Only W and Rz gates are supported, found {type(operation)}.__name__"

            if isinstance(operation, schema.GlobalW):
                self.lower_w_gates(
                    tuple(self.qubit_id_map.keys()), operation.theta, operation.phi
                )
            elif isinstance(operation, schema.LocalW):
                self.lower_w_gates(
                    operation.participants, operation.theta, operation.phi
                )
            elif isinstance(operation, schema.GlobalRz):
                self.lower_rz_gates(tuple(self.qubit_id_map.keys()), operation.phi)
            elif isinstance(operation, schema.LocalRz):
                self.lower_rz_gates(operation.participants, operation.phi)
            elif isinstance(operation, schema.Measurement):
                self.lower_measurement(operation)

    def process_cz_pauli_error(
        self,
        participants: Tuple[Tuple[int] | Tuple[int, int], ...],
        node: schema.CZError[schema.PauliErrorModel],
    ):

        storage_error = node.storage_error
        single_error = node.single_error
        entangled_error = node.entangled_error

        assert isinstance(
            storage_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"
        assert isinstance(
            single_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"
        assert isinstance(
            entangled_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"

        self.lower_pauli_errors(storage_error)

        single_error_dict = dict(single_error.errors)
        entangled_error_dict = dict(entangled_error.errors)

        single_layers = {}
        paired_layers = {}
        unpaired_layers = {}

        for participant in participants:
            if len(participant) == 1:
                p_single = single_error_dict[participant[0]]
                single_layers.setdefault(p_single, []).append(participant[0])
            elif len(participant) == 2:
                p_ctrl = entangled_error_dict[participant[0]]
                up_ctrl = single_error_dict[participant[0]]
                p_qarg = entangled_error_dict[participant[1]]
                up_qarg = single_error_dict[participant[1]]
                paired_layers.setdefault((p_ctrl, p_qarg), []).append(participant)
                unpaired_layers.setdefault((up_ctrl, up_qarg), []).append(participant)

        for (px, py, pz), qubits in single_layers.items():
            self.block_list.append(
                qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in qubits))
            )
            self.block_list.append(
                noise.native.PauliChannel(px=px, py=py, pz=pz, qargs=qargs.result)
            )

        for (p_ctrl, p_qarg), qubits in paired_layers.items():
            ctrls, qargs = list(zip(*qubits))
            self.block_list.append(
                ctrls := ilist.New(values=tuple(self.qubit_id_map[q] for q in ctrls))
            )
            self.block_list.append(
                qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in qargs))
            )
            self.block_list.append(
                noise.native.CZPauliChannel(
                    paired=True,
                    px_ctrl=p_ctrl[0],
                    py_ctrl=p_ctrl[1],
                    pz_ctrl=p_ctrl[2],
                    px_qarg=p_qarg[0],
                    py_qarg=p_qarg[1],
                    pz_qarg=p_qarg[2],
                    ctrls=ctrls.result,
                    qargs=qargs.result,
                )
            )

        for (p_ctrl, p_qarg), qubits in unpaired_layers.items():
            ctrls, qargs = list(zip(*qubits))
            self.block_list.append(
                ctrls := ilist.New(values=tuple(self.qubit_id_map[q] for q in ctrls))
            )
            self.block_list.append(
                qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in qargs))
            )
            self.block_list.append(
                noise.native.CZPauliChannel(
                    paired=False,
                    px_ctrl=p_ctrl[0],
                    py_ctrl=p_ctrl[1],
                    pz_ctrl=p_ctrl[2],
                    px_qarg=p_qarg[0],
                    py_qarg=p_qarg[1],
                    pz_qarg=p_qarg[2],
                    ctrls=ctrls.result,
                    qargs=qargs.result,
                )
            )

    def lower_cz_gates(self, node: schema.CZ):
        ctrls, qargs = list(zip(*(p for p in node.participants if len(p) == 2)))
        self.block_list.append(
            ctrls := ilist.New(values=tuple(self.qubit_id_map[q] for q in ctrls))
        )
        self.block_list.append(
            qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in qargs))
        )
        self.block_list.append(parallel.CZ(ctrls=ctrls.result, qargs=qargs.result))

    def lower_w_gates(self, participants: Sequence[int], theta: float, phi: float):
        self.block_list.append(
            qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in participants))
        )
        self.block_list.append(
            parallel.UGate(
                theta=self.lower_full_turns(theta),
                phi=self.lower_full_turns(phi + 0.5),
                lam=self.lower_full_turns(-(0.5 + phi)),
                qargs=qargs.result,
            )
        )

    def lower_rz_gates(self, participants: Tuple[int, ...], phi: float):
        self.block_list.append(
            qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in participants))
        )
        self.block_list.append(
            parallel.RZ(theta=self.lower_full_turns(phi), qargs=qargs.result)
        )

    def lower_pauli_errors(self, operator_error: schema.PauliErrorModel):
        assert isinstance(
            operator_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"

        layers = {}

        for qubit_num, pauli_errors in operator_error.errors:
            layers.setdefault(pauli_errors, []).append(qubit_num)

        for (px, py, pz), qubits in layers.items():
            self.block_list.append(
                qargs := ilist.New(values=tuple(self.qubit_id_map[q] for q in qubits))
            )
            self.block_list.append(
                noise.native.PauliChannel(px=px, py=py, pz=pz, qargs=qargs.result)
            )

    def lower_measurement(self, operation: schema.Measurement):
        for participant in operation.participants:
            qubit = self.qubit_id_map[participant]
            bit = self.bit_id_map[participant]
            self.block_list.append(qasm2.core.Measure(qarg=qubit, carg=bit))

    def lower_atom_loss(self, survival_probs: Tuple[float, ...]):
        layers = {}

        for qubit_num, survival_prob in zip(self.qubit_list, survival_probs):
            layers.setdefault(survival_prob, []).append(qubit_num)

        for survival_prob, qubits in layers.items():
            self.block_list.append(qargs := ilist.New(values=qubits))
            self.block_list.append(
                noise.native.AtomLossChannel(prob=survival_prob, qargs=qargs.result)
            )

    def lower_number(self, value: float | int) -> ir.SSAValue:
        if isinstance(value, int):
            stmt = qasm2.expr.ConstInt(value=value)
        else:
            stmt = qasm2.expr.ConstFloat(value=value)

        self.block_list.append(stmt)
        return stmt.result

    def lower_full_turns(self, value: float) -> ir.SSAValue:
        const_pi = qasm2.expr.ConstPI()
        self.block_list.append(const_pi)
        turns = self.lower_number(2 * value)
        mul = qasm2.expr.Mul(const_pi.result, turns)
        self.block_list.append(mul)
        return mul.result



from kirin import ir, types
from kirin.decl import info, statement
from kirin.dialects import ilist
from bloqade.qasm2.types import QubitType

from ._dialect import dialect


@statement(dialect=dialect)
class PauliChannel(ir.Statement):

    traits = frozenset({ir.FromPythonCall()})

    px: float = info.attribute(types.Float)
    py: float = info.attribute(types.Float)
    pz: float = info.attribute(types.Float)
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType])


NumQubits = types.TypeVar("NumQubits")


@statement(dialect=dialect)
class CZPauliChannel(ir.Statement):

    traits = frozenset({ir.FromPythonCall()})

    paired: bool = info.attribute(types.Bool)
    px_ctrl: float = info.attribute(types.Float)
    py_ctrl: float = info.attribute(types.Float)
    pz_ctrl: float = info.attribute(types.Float)
    px_qarg: float = info.attribute(types.Float)
    py_qarg: float = info.attribute(types.Float)
    pz_qarg: float = info.attribute(types.Float)
    ctrls: ir.SSAValue = info.argument(ilist.IListType[QubitType, NumQubits])
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType, NumQubits])


@statement(dialect=dialect)
class AtomLossChannel(ir.Statement):

    traits = frozenset({ir.FromPythonCall()})

    prob: float = info.attribute(types.Float)
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType])



from typing import Any

from bloqade.noise import native
from kirin.dialects import ilist
from kirin.lowering import wraps
from bloqade.qasm2.types import Qubit


@wraps(native.AtomLossChannel)
def atom_loss_channel(
    qargs: ilist.IList[Qubit, Any] | list, *, prob: float
) -> None: ...


@wraps(native.PauliChannel)
def pauli_channel(
    qargs: ilist.IList[Qubit, Any] | list, *, px: float, py: float, pz: float
) -> None: ...


@wraps(native.CZPauliChannel)
def cz_pauli_channel(
    ctrls: ilist.IList[Qubit, Any] | list,
    qarg2: ilist.IList[Qubit, Any] | list,
    *,
    px_ctrl: float,
    py_ctrl: float,
    pz_ctrl: float,
    px_qarg: float,
    py_qarg: float,
    pz_qarg: float,
    paired: bool,
) -> None: ...



import math

from bloqade.noise import native


def test_zone_model_deconflict():
    noise_model = native.TwoRowZoneModel()

    result = noise_model.deconflict([0, 4, 2, 1], [3, 5, 6, 7])
    assert result == [((0, 1), (3, 7)), ((2,), (6,)), ((4,), (5,))]


def test_zone_model_assign_slots():
    noise_model = native.TwoRowZoneModel()

    result = noise_model.assign_gate_slots([0, 1, 5], [3, 7, 9])
    assert result == {1: (1, 7), 0: (0, 3), 2: (5, 9)}


def test_move_duration():
    noise_model = native.TwoRowZoneModel()

    slots = {1: (1, 7), 0: (0, 3), 2: (5, 9)}
    result = noise_model.calculate_move_duration(slots)
    assert math.isclose(result, 99.14523415238915)


def test_coeff():

    noise_model = native.TwoRowZoneModel()
    result = noise_model.parallel_cz_errors([0, 4, 2, 1], [3, 5, 6, 7], [])

    qubit_result = {}

    for p, qubits in result.items():
        for qubit in qubits:
            qubit_result[qubit] = p

    expected_p = (
        0.0012829622129098485,
        0.0012829622129098485,
        0.0012829622129098485,
        0.00038347256559846324,
    )

    expected_qubit_result = {
        0: expected_p,
        1: expected_p,
        2: expected_p,
        3: expected_p,
        4: expected_p,
        5: expected_p,
        6: expected_p,
        7: expected_p,
    }

    for q, p in qubit_result.items():
        assert math.isclose(p[0], expected_qubit_result[q][0])
        assert math.isclose(p[1], expected_qubit_result[q][1])
        assert math.isclose(p[2], expected_qubit_result[q][2])
        assert math.isclose(p[3], expected_qubit_result[q][3])


if __name__ == "__main__":
    test_zone_model_assign_slots()
    test_move_duration()



from kirin import ir, types
from bloqade import qasm2
from bloqade.noise import native
from kirin.dialects import func, ilist
from bloqade.analysis import address
from kirin.dialects.py import constant
from bloqade.test_utils import assert_nodes
from bloqade.qasm2.dialects import uop, core, glob, parallel
from bloqade.qasm2.passes.noise import NoisePass
from bloqade.qasm2.rewrite.heuristic_noise import NoiseRewriteRule


class TestNoise(native.MoveNoiseModelABC):
    def parallel_cz_errors(cls, ctrls, qargs, rest):
        return {(0.01, 0.01, 0.01, 0.01): ctrls + qargs + rest}


def test_single_qubit_noise():
    px = 0.01
    py = 0.01
    pz = 0.01
    p_loss = 0.01

    noise_params = native.GateNoiseParams(
        local_px=px, local_py=py, local_pz=pz, local_loss_prob=p_loss
    )
    model = TestNoise()

    test_qubit = ir.TestValue(type=qasm2.QubitType)
    address_analysis = {test_qubit: address.AddressQubit(0)}
    rule = NoiseRewriteRule(address_analysis, noise_params, model)
    rule.qubit_ssa_value[0] = test_qubit
    block = ir.Block(
        [
            stmt := uop.X(qarg=test_qubit),
        ]
    )

    rule.rewrite(stmt)

    expected_block = ir.Block(
        [
            qubit_list := ilist.New(values=[test_qubit]),
            native.PauliChannel(qargs=qubit_list.result, px=px, py=py, pz=pz),
            native.AtomLossChannel(qargs=qubit_list.result, prob=p_loss),
            stmt.from_stmt(stmt),
        ]
    )

    assert_nodes(block, expected_block)


def test_parallel_qubit_noise():
    px = 0.01
    py = 0.01
    pz = 0.01
    p_loss = 0.01

    noise_params = native.GateNoiseParams(
        local_px=px, local_py=py, local_pz=pz, local_loss_prob=p_loss
    )
    model = TestNoise()

    test_qubit = ir.TestValue(type=qasm2.QubitType)
    qubit_list = ilist.New(values=[test_qubit])

    test_float = ir.TestValue(type=types.Float)
    address_analysis = {
        test_qubit: address.AddressQubit,
        qubit_list.result: address.AddressTuple([address.AddressQubit(0)]),
    }
    rule = NoiseRewriteRule(address_analysis, noise_params, model)
    rule.qubit_ssa_value[0] = test_qubit
    block = ir.Block(
        [
            qubit_list,
            stmt := parallel.UGate(
                qargs=qubit_list.result,
                theta=test_float,
                phi=test_float,
                lam=test_float,
            ),
        ]
    )

    rule.rewrite(stmt)

    expected_block = ir.Block(
        [
            qubit_list := qubit_list.from_stmt(qubit_list),
            native.PauliChannel(qargs=qubit_list.result, px=px, py=py, pz=pz),
            native.AtomLossChannel(qargs=qubit_list.result, prob=p_loss),
            parallel.UGate(
                qargs=qubit_list.result,
                theta=test_float,
                phi=test_float,
                lam=test_float,
            ),
        ]
    )

    assert_nodes(block, expected_block)


def test_cz_gate_noise():
    px = 0.01
    py = 0.01
    pz = 0.01
    p_loss = 0.01

    noise_params = native.GateNoiseParams(
        cz_paired_gate_px=px,
        cz_paired_gate_py=py,
        cz_paired_gate_pz=pz,
        cz_gate_loss_prob=p_loss,
        cz_unpaired_gate_px=px,
        cz_unpaired_gate_py=py,
        cz_unpaired_gate_pz=pz,
        cz_unpaired_loss_prob=p_loss,
    )

    model = TestNoise()

    ctrl_qubit = ir.TestValue(type=qasm2.QubitType)
    qarg_qubit = ir.TestValue(type=qasm2.QubitType)
    address_analysis = {
        ctrl_qubit: address.AddressQubit(0),
        qarg_qubit: address.AddressQubit(1),
    }
    rule = NoiseRewriteRule(address_analysis, noise_params, model)
    rule.qubit_ssa_value[0] = ctrl_qubit
    rule.qubit_ssa_value[1] = qarg_qubit
    block = ir.Block(
        [
            stmt := uop.CZ(qarg=qarg_qubit, ctrl=ctrl_qubit),
        ]
    )

    rule.rewrite(stmt)

    expected_block = ir.Block(
        [
            ctrls := ilist.New(values=[ctrl_qubit]),
            qargs := ilist.New(values=[qarg_qubit]),
            all_qubits := ilist.New(values=[ctrl_qubit, qarg_qubit]),
            native.AtomLossChannel(qargs=all_qubits.result, prob=p_loss),
            native.PauliChannel(qargs=all_qubits.result, px=px, py=py, pz=pz),
            native.CZPauliChannel(
                ctrls=ctrls.result,
                qargs=qargs.result,
                px_ctrl=px,
                py_ctrl=py,
                pz_ctrl=pz,
                px_qarg=px,
                py_qarg=py,
                pz_qarg=pz,
                paired=True,
            ),
            native.CZPauliChannel(
                ctrls=ctrls.result,
                qargs=qargs.result,
                px_ctrl=px,
                py_ctrl=py,
                pz_ctrl=pz,
                px_qarg=px,
                py_qarg=py,
                pz_qarg=pz,
                paired=False,
            ),
            native.AtomLossChannel(qargs=ctrls.result, prob=p_loss),
            native.AtomLossChannel(qargs=qargs.result, prob=p_loss),
            stmt.from_stmt(stmt),
        ]
    )

    assert_nodes(block, expected_block)


def test_parallel_cz_gate_noise():
    px = 0.01
    py = 0.01
    pz = 0.01
    p_loss = 0.01

    noise_params = native.GateNoiseParams(
        cz_paired_gate_px=px,
        cz_paired_gate_py=py,
        cz_paired_gate_pz=pz,
        cz_gate_loss_prob=p_loss,
        cz_unpaired_gate_px=px,
        cz_unpaired_gate_py=py,
        cz_unpaired_gate_pz=pz,
        cz_unpaired_loss_prob=p_loss,
    )

    model = TestNoise()

    ctrl_qubit = ir.TestValue(type=qasm2.QubitType)
    qarg_qubit = ir.TestValue(type=qasm2.QubitType)
    ctrl_list = ilist.New(values=[ctrl_qubit])
    qarg_list = ilist.New(values=[qarg_qubit])
    address_analysis = {
        ctrl_qubit: address.AddressQubit(0),
        qarg_qubit: address.AddressQubit(1),
        ctrl_list.result: address.AddressTuple([address.AddressQubit(0)]),
        qarg_list.result: address.AddressTuple([address.AddressQubit(1)]),
    }
    rule = NoiseRewriteRule(address_analysis, noise_params, model)
    rule.qubit_ssa_value[0] = ctrl_qubit
    rule.qubit_ssa_value[1] = qarg_qubit
    block = ir.Block(
        [
            ctrl_list,
            qarg_list,
            stmt := parallel.CZ(qargs=qarg_list.result, ctrls=ctrl_list.result),
        ]
    )

    rule.rewrite(stmt)

    expected_block = ir.Block(
        [
            ctrl_list := ctrl_list.from_stmt(ctrl_list),
            qarg_list := qarg_list.from_stmt(qarg_list),
            all_qubits := ilist.New(values=[ctrl_qubit, qarg_qubit]),
            native.AtomLossChannel(qargs=all_qubits.result, prob=p_loss),
            native.PauliChannel(qargs=all_qubits.result, px=px, py=py, pz=pz),
            native.CZPauliChannel(
                ctrls=ctrl_list.result,
                qargs=qarg_list.result,
                px_ctrl=px,
                py_ctrl=py,
                pz_ctrl=pz,
                px_qarg=px,
                py_qarg=py,
                pz_qarg=pz,
                paired=True,
            ),
            native.CZPauliChannel(
                ctrls=ctrl_list.result,
                qargs=qarg_list.result,
                px_ctrl=px,
                py_ctrl=py,
                pz_ctrl=pz,
                px_qarg=px,
                py_qarg=py,
                pz_qarg=pz,
                paired=False,
            ),
            native.AtomLossChannel(qargs=ctrl_list.result, prob=p_loss),
            native.AtomLossChannel(qargs=qarg_list.result, prob=p_loss),
            parallel.CZ(qargs=qarg_list.result, ctrls=ctrl_list.result),
        ]
    )

    assert_nodes(block, expected_block)


def test_global_noise():

    @qasm2.extended
    def test_method():
        q0 = qasm2.qreg(1)
        q1 = qasm2.qreg(1)
        glob.UGate([q0, q1], 0.1, 0.2, 0.3)

    px = 0.01
    py = 0.01
    pz = 0.01
    p_loss = 0.01

    noise_params = native.GateNoiseParams(
        global_loss_prob=p_loss, global_px=px, global_py=py, global_pz=pz
    )

    model = TestNoise()

    test_method.print()

    NoisePass(test_method.dialects, noise_model=model, gate_noise_params=noise_params)(
        test_method
    )

    expected_block = ir.Block(
        [
            n_qubits := constant.Constant(1),
            reg0 := core.QRegNew(n_qubits.result),
            zero := constant.Constant(0),
            q0 := core.QRegGet(reg0.result, zero.result),
            reg1 := core.QRegNew(n_qubits.result),
            q1 := core.QRegGet(reg1.result, zero.result),
            reg_list := ilist.New(values=[reg0.result, reg1.result]),
            theta := constant.Constant(0.1),
            phi := constant.Constant(0.2),
            lam := constant.Constant(0.3),
            qargs := ilist.New(values=[q0.result, q1.result]),
            native.PauliChannel(qargs.result, px=px, py=py, pz=pz),
            native.AtomLossChannel(qargs.result, prob=p_loss),
            glob.UGate(reg_list.result, theta.result, phi.result, lam.result),
            none := func.ConstantNone(),
            func.Return(none.result),
        ],
        argtypes=[types.MethodType[[], types.NoneType]],
    )

    expected_block.args[0].name = "test_method_self"

    reg0.result.name = "q0"
    reg1.result.name = "q1"
    expected_region = ir.Region([expected_block])
    assert_nodes(test_method.callable_region, expected_region)



from .fold import QASM2Fold as QASM2Fold
from .noise import NoisePass as NoisePass
from .py2qasm import Py2QASM as Py2QASM
from .qasm2py import QASM2Py as QASM2Py
from .parallel import UOpToParallel as UOpToParallel



from typing import Dict, List
from dataclasses import dataclass

from kirin import ir
from bloqade import qasm2
from kirin.rewrite import abc, result
from kirin.dialects import py, ilist
from bloqade.analysis import address
from bloqade.qasm2.dialects import glob


@dataclass
class GlobalRewriteBase:
    address_analysis: Dict[ir.SSAValue, address.Address]

    def get_qubit_ssa(self, node: glob.UGate):
        new_stmts: List[ir.Statement] = []
        qubit_ssa: List[ir.SSAValue] = []
        # can't rewrite if the registers are coming from a block argument
        if not isinstance(node.registers, ir.ResultValue):
            return new_stmts, None

        if not isinstance(node.registers.owner, ilist.New):
            return new_stmts, None

        register_ssa_values = node.registers.owner.values

        for register_ssa in register_ssa_values:
            addr = self.address_analysis.get(register_ssa, address.Address.top())
            if not isinstance(addr, address.AddressReg):
                new_stmts.clear()
                return new_stmts, None

            for qubit in range(len(addr.data)):
                new_stmts.append(idx_stmt := py.constant.Constant(value=qubit))
                new_stmts.append(
                    qubit_stmt := qasm2.core.QRegGet(
                        reg=register_ssa, idx=idx_stmt.result
                    )
                )
                qubit_ssa.append(qubit_stmt.result)

        return new_stmts, qubit_ssa


@dataclass
class GlobalToParallelRule(abc.RewriteRule, GlobalRewriteBase):

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if type(node) in glob.dialect.stmts:
            return getattr(self, f"rewrite_{node.name}")(node)

        return result.RewriteResult()

    def rewrite_ugate(self, node: glob.UGate):

        new_stmts, qubit_ssa = self.get_qubit_ssa(node)

        if qubit_ssa is None:
            return result.RewriteResult()

        new_stmts.append(qargs := ilist.New(values=qubit_ssa))
        new_stmts.append(
            qasm2.dialects.parallel.UGate(
                qargs=qargs.result, theta=node.theta, phi=node.phi, lam=node.lam
            )
        )

        for stmt in new_stmts:
            stmt.insert_before(node)

        node.delete()

        return result.RewriteResult(has_done_something=True)


@dataclass
class GlobalToUOpRule(abc.RewriteRule, GlobalRewriteBase):

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if type(node) in glob.dialect.stmts:
            return getattr(self, f"rewrite_{node.name}")(node)

        return result.RewriteResult()

    def rewrite_ugate(self, node: glob.UGate):

        new_stmts, qubit_ssa = self.get_qubit_ssa(node)

        if qubit_ssa is None:
            return result.RewriteResult()

        for qarg in qubit_ssa:
            new_stmts.append(
                qasm2.uop.UGate(qarg=qarg, theta=node.theta, phi=node.phi, lam=node.lam)
            )

        for stmt in new_stmts:
            stmt.insert_before(node)

        node.delete()
        return result.RewriteResult(has_done_something=True)



import math
from typing import List, Optional
from functools import cached_property
from dataclasses import field, dataclass

import cirq
import cirq.transformers
import cirq.contrib.qasm_import
import cirq.transformers.target_gatesets
import cirq.transformers.target_gatesets.compilation_target_gateset
from kirin import ir
from kirin.rewrite import abc, result
from kirin.dialects import py
from bloqade.qasm2.dialects import uop, expr
from cirq.circuits.qasm_output import QasmUGate
from cirq.transformers.target_gatesets.compilation_target_gateset import (
    CompilationTargetGateset,
)


# rydeberg gate sets
class RydbergTargetGateset(cirq.CZTargetGateset):
    def __init__(self, *, cnz_max_size: int = 2, atol: float = 1e-8):
        additional = [cirq.Z.controlled(cn) for cn in range(2, cnz_max_size)]
        super().__init__(atol=atol, additional_gates=additional)
        self.cnz_max_size = cnz_max_size

    @property
    def num_qubits(self) -> int:
        return max(2, self.cnz_max_size)


# decompose the CU by defining a custom Cirq Gate with the qelib1 definition
# Need to be careful about the fact that U(theta, phi, lambda) in standard QASM2
# and its variants
class CU(cirq.Gate):
    def __init__(self, theta, phi, lam, gamma):
        super(CU, self)
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma

    def _num_qubits_(self):
        return 2

    def _decompose_(self, qubits):
        ctrl, target = qubits
        # taken from qelib1 definition
        # p(gamma) c;
        yield QasmUGate(0, 0, self.gamma / math.pi)(ctrl)
        # p((lambda+phi/2)) c;
        yield QasmUGate(0, 0, ((self.lam + self.phi) / 2) / math.pi)(ctrl)
        # p((lambda-phi/2)) t;
        yield QasmUGate(0, 0, ((self.lam - self.phi) / 2) / math.pi)(target)
        # cx c,t
        yield cirq.CX(ctrl, target)
        # u(-theta/2, 0, -(phi+lambda/2)) t;
        yield QasmUGate(
            (-self.theta / 2) / math.pi, 0, (-(self.phi + self.lam) / 2) / math.pi
        )(target)
        # cx c,t
        yield cirq.CX(ctrl, target)
        # u(theta/2, phi, 0) t;
        yield QasmUGate((self.theta / 2) / math.pi, self.phi / math.pi, 0)(target)

    def _circuit_diagram_info_(self, args):
        return "*", "CU"


def one_qubit_gate_to_u3_angles(op: cirq.Operation) -> tuple[float, float, float]:
    lam, theta, phi = (  # Z angle, Y angle, then Z angle
        cirq.deconstruct_single_qubit_matrix_into_angles(cirq.unitary(op))
    )

    return theta, phi, lam


@dataclass
class RydbergGateSetRewriteRule(abc.RewriteRule):
    # NOTE
    # 1. this can only rewrite qasm2.main and qasm2.gate!
    dialect_group: ir.DialectGroup
    gateset: CompilationTargetGateset = field(default_factory=RydbergTargetGateset)

    @cached_property
    def cached_qubits(self) -> tuple[cirq.LineQubit, ...]:

        # qasm2 stmts only have up to 3 qubits gates, so we cached only 3.
        return tuple(cirq.LineQubit(i) for i in range(3))

    @cached_property
    def const_float_type(self):
        if expr.dialect in self.dialect_group.data:
            return expr.ConstFloat
        else:
            return py.constant.Constant

    def const_float(self, value: float):
        return self.const_float_type(value=value)

    @cached_property
    def const_pi(self):
        if expr in self.dialect_group.data:
            return expr.ConstPI()
        else:
            return py.constant.Constant(value=math.pi)

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        # only deal with uop
        if type(node) in uop.dialect.stmts:
            return getattr(self, f"rewrite_{node.name}")(node)

        return result.RewriteResult()

    def rewrite_barrier(self, node: uop.Barrier) -> result.RewriteResult:
        return result.RewriteResult()

    def rewrite_cz(self, node: uop.CZ) -> result.RewriteResult:
        return result.RewriteResult()

    def rewrite_CX(self, node: uop.CX) -> result.RewriteResult:
        return self._rewrite_2q_ctrl_gates(
            cirq.CX(self.cached_qubits[0], self.cached_qubits[1]), node
        )

    def rewrite_cy(self, node: uop.CY) -> result.RewriteResult:
        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(cirq.Y, 1)(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

    def rewrite_U(self, node: uop.UGate) -> result.RewriteResult:
        return result.RewriteResult()

    def rewrite_id(self, node: uop.Id) -> result.RewriteResult:
        node.delete()  # just delete the identity gate
        return result.RewriteResult(has_done_something=True)

    def rewrite_h(self, node: uop.H) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.H(self.cached_qubits[0]), node)

    def rewrite_x(self, node: uop.X) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.X(self.cached_qubits[0]), node)

    def rewrite_y(self, node: uop.Y) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.Y(self.cached_qubits[0]), node)

    def rewrite_z(self, node: uop.Z) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.Z(self.cached_qubits[0]), node)

    def rewrite_s(self, node: uop.S) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.S(self.cached_qubits[0]), node)

    def rewrite_sdg(self, node: uop.Sdag) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.S(self.cached_qubits[0]) ** -1, node)

    def rewrite_t(self, node: uop.T) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.T(self.cached_qubits[0]), node)

    def rewrite_tdg(self, node: uop.Tdag) -> result.RewriteResult:
        return self._rewrite_1q_gates(cirq.T(self.cached_qubits[0]) ** -1, node)

    def rewrite_sx(self, node: uop.SX) -> result.RewriteResult:
        return self._rewrite_1q_gates(
            cirq.XPowGate(exponent=0.5).on(self.cached_qubits[0]), node
        )

    def rewrite_sxdg(self, node: uop.SXdag) -> result.RewriteResult:
        return self._rewrite_1q_gates(
            cirq.XPowGate(exponent=-0.5).on(self.cached_qubits[0]), node
        )

    def rewrite_u1(self, node: uop.U1) -> result.RewriteResult:
        theta = node.lam
        (phi := self.const_float(value=0.0)).insert_before(node)
        node.replace_by(
            uop.UGate(qarg=node.qarg, theta=phi.result, phi=phi.result, lam=theta)
        )
        return result.RewriteResult(has_done_something=True)

    def rewrite_u2(self, node: uop.U2) -> result.RewriteResult:
        phi = node.phi
        lam = node.lam
        (theta := self.const_float(value=math.pi / 2)).insert_before(node)
        node.replace_by(uop.UGate(qarg=node.qarg, theta=theta.result, phi=phi, lam=lam))
        return result.RewriteResult(has_done_something=True)

    def rewrite_rx(self, node: uop.RX) -> result.RewriteResult:
        theta = node.theta
        (phi := self.const_float(value=math.pi / 2)).insert_before(node)
        (lam := self.const_float(value=-math.pi / 2)).insert_before(node)
        node.replace_by(
            uop.UGate(qarg=node.qarg, theta=theta, phi=phi.result, lam=lam.result)
        )
        return result.RewriteResult(has_done_something=True)

    def rewrite_ry(self, node: uop.RY) -> result.RewriteResult:
        theta = node.theta
        (phi := self.const_float(value=0.0)).insert_before(node)
        node.replace_by(
            uop.UGate(qarg=node.qarg, theta=theta, phi=phi.result, lam=phi.result)
        )
        return result.RewriteResult(has_done_something=True)

    def rewrite_rz(self, node: uop.RZ) -> result.RewriteResult:
        theta = node.theta
        (phi := self.const_float(value=0.0)).insert_before(node)
        node.replace_by(
            uop.UGate(qarg=node.qarg, theta=phi.result, phi=phi.result, lam=theta)
        )
        return result.RewriteResult(has_done_something=True)

    def rewrite_crx(self, node: uop.CRX) -> result.RewriteResult:
        lam = self._get_const_value(node.lam)

        if lam is None:
            return result.RewriteResult()

        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(cirq.Rx(rads=lam), 1).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

    def rewrite_cry(self, node: uop.CRY) -> result.RewriteResult:
        lam = self._get_const_value(node.lam)

        if lam is None:
            return result.RewriteResult()

        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(cirq.Ry(rads=lam), 1).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

    def rewrite_crz(self, node: uop.CRZ) -> result.RewriteResult:
        lam = self._get_const_value(node.lam)

        if lam is None:
            return result.RewriteResult()

        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(cirq.Rz(rads=lam), 1).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

    def rewrite_cu1(self, node: uop.CU1) -> result.RewriteResult:

        lam = self._get_const_value(node.lam)

        if lam is None:
            return result.RewriteResult()

        # cirq.ControlledGate(u3(0, 0, lambda))
        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(QasmUGate(0, 0, lam / math.pi)).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )
        pass

    def rewrite_cu3(self, node: uop.CU3) -> result.RewriteResult:

        theta = self._get_const_value(node.theta)
        lam = self._get_const_value(node.lam)
        phi = self._get_const_value(node.phi)

        if not all((theta, phi, lam)):
            return result.RewriteResult()

        # cirq.ControlledGate(u3(theta, lambda phi))
        return self._rewrite_2q_ctrl_gates(
            cirq.ControlledGate(
                QasmUGate(theta / math.pi, phi / math.pi, lam / math.pi)
            ).on(self.cached_qubits[0], self.cached_qubits[1]),
            node,
        )

    def rewrite_cu(self, node: uop.CU) -> result.RewriteResult:

        gamma = self._get_const_value(node.gamma)
        theta = self._get_const_value(node.theta)
        lam = self._get_const_value(node.lam)
        phi = self._get_const_value(node.phi)

        # need to create custom 2q gate, then feed that into rewrite_2q

        return self._rewrite_2q_ctrl_gates(
            CU(theta, phi, lam, gamma).on(self.cached_qubits[0], self.cached_qubits[1]),
            node,
        )

    def rewrite_rxx(self, node: uop.RXX) -> result.RewriteResult:

        theta = self._get_const_value(node.theta)

        if theta is None:
            return result.RewriteResult()

        # even though the XX gate is not controlled,
        # the end U + CZ decomposition that happens internally means
        return self._rewrite_2q_ctrl_gates(
            cirq.XXPowGate(exponent=theta / math.pi).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

    def rewrite_rzz(self, node: uop.RZZ) -> result.RewriteResult:
        theta = self._get_const_value(node.theta)

        if theta is None:
            return result.RewriteResult()

        return self._rewrite_2q_ctrl_gates(
            cirq.ZZPowGate(exponent=theta / math.pi).on(
                self.cached_qubits[0], self.cached_qubits[1]
            ),
            node,
        )

        """
        return self._rewrite_2q_ctrl_gates(
            cirq.ZZPowGate(exponent = theta).on(self.cached_qubits[0], self.cached_qubits[1])
            ,node
        )
        """

    def rewrite_swap(self, node: uop.Swap):
        return self._rewrite_2q_ctrl_gates(
            cirq.SWAP(self.cached_qubits[0], self.cached_qubits[1]), node
        )

    def _get_const_value(self, ssa: ir.SSAValue) -> Optional[float | int]:
        if not isinstance(ssa, ir.ResultValue):
            return None

        match ssa.owner:
            case expr.ConstFloat(value=value):
                return value
            case expr.ConstInt(value=value):
                return value
            case py.constant.Constant(value=float() as value) | py.constant.Constant(
                value=int() as value
            ):
                return value
            case expr.ConstPI():
                return math.pi
            case _:
                return None

    def _generate_1q_gate_stmts(self, cirq_gate: cirq.Operation, qarg: ir.SSAValue):
        target_gates = self.gateset.decompose_to_target_gateset(cirq_gate, 0)

        if isinstance(target_gates, cirq.GateOperation):
            target_gates = [target_gates]

        new_stmts = []
        for new_gate in target_gates:
            theta, phi, lam = one_qubit_gate_to_u3_angles(new_gate)
            theta_stmt = self.const_float(value=theta)
            phi_stmt = self.const_float(value=phi)
            lam_stmt = self.const_float(value=lam)

            new_stmts.append(theta_stmt)
            new_stmts.append(phi_stmt)
            new_stmts.append(lam_stmt)
            new_stmts.append(
                uop.UGate(
                    qarg=qarg,
                    theta=theta_stmt.result,
                    phi=phi_stmt.result,
                    lam=lam_stmt.result,
                )
            )
        return new_stmts

    def _rewrite_1q_gates(
        self, cirq_gate: cirq.Operation, node: uop.SingleQubitGate
    ) -> result.RewriteResult:
        new_gate_stmts = self._generate_1q_gate_stmts(cirq_gate, node.qarg)
        return self._rewrite_gate_stmts(new_gate_stmts, node)

    def _generate_2q_ctrl_gate_stmts(
        self, cirq_gate: cirq.Operation, qubits_ssa: List[ir.SSAValue]
    ) -> list[ir.Statement]:
        target_gates = self.gateset.decompose_to_target_gateset(cirq_gate, 0)
        new_stmts = []
        for new_gate in target_gates:
            if len(new_gate.qubits) == 1:
                # 1q
                phi0, phi1, phi2 = one_qubit_gate_to_u3_angles(new_gate)
                phi0_stmt = self.const_float(value=phi0)
                phi1_stmt = self.const_float(value=phi1)
                phi2_stmt = self.const_float(value=phi2)

                new_stmts.append(phi0_stmt)
                new_stmts.append(phi1_stmt)
                new_stmts.append(phi2_stmt)
                new_stmts.append(
                    uop.UGate(
                        qarg=qubits_ssa[new_gate.qubits[0].x],
                        theta=phi0_stmt.result,
                        phi=phi1_stmt.result,
                        lam=phi2_stmt.result,
                    )
                )
            else:
                # 2q
                new_stmts.append(uop.CZ(ctrl=qubits_ssa[0], qarg=qubits_ssa[1]))

        return new_stmts

    def _rewrite_2q_ctrl_gates(
        self, cirq_gate: cirq.Operation, node: uop.TwoQubitCtrlGate
    ) -> result.RewriteResult:
        new_gate_stmts = self._generate_2q_ctrl_gate_stmts(
            cirq_gate, [node.ctrl, node.qarg]
        )
        return self._rewrite_gate_stmts(new_gate_stmts, node)

    def _rewrite_gate_stmts(
        self, new_gate_stmts: list[ir.Statement], node: ir.Statement
    ):

        node.replace_by(new_gate_stmts[0])
        node = new_gate_stmts[0]

        for stmt in new_gate_stmts[1:]:
            stmt.insert_after(node)
            node = stmt

        return result.RewriteResult(has_done_something=True)



import io
import difflib

from kirin import ir, print as pprint
from rich.console import Console


def print_diff(node: pprint.Printable, expected_node: pprint.Printable):
    gn_con = Console(record=True, file=io.StringIO())
    node.print(console=gn_con)

    expected_con = Console(record=True, file=io.StringIO())
    expected_node.print(console=expected_con)

    expected = expected_con.export_text()

    gn = gn_con.export_text()
    diff = difflib.Differ().compare(
        expected.splitlines(),
        gn.splitlines(),
    )

    print("\n".join(diff))


def assert_nodes(node: ir.IRNode, expected_node: ir.IRNode):
    try:
        assert node.is_equal(expected_node)
    except AssertionError as e:
        print_diff(node, expected_node)
        raise e


def assert_methods(mt: ir.Method, expected_mt: ir.Method):
    assert_nodes(mt.code, expected_mt.code)



from dataclasses import field, dataclass

from kirin import ir
from kirin.passes import Pass, TypeInfer
from kirin.rewrite import (
    Walk,
    Chain,
    Inline,
    Fixpoint,
    WrapConst,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    InlineGetField,
    DeadCodeElimination,
    CommonSubexpressionElimination,
)
from kirin.analysis import const
from kirin.dialects import scf, ilist
from kirin.ir.method import Method
from kirin.rewrite.abc import RewriteResult
from bloqade.qasm2.dialects import expr


@dataclass
class QASM2Fold(Pass):
    """Fold pass for qasm2.extended"""

    constprop: const.Propagate = field(init=False)
    inline_gate_subroutine: bool = True

    def __post_init__(self):
        self.constprop = const.Propagate(self.dialects)
        self.typeinfer = TypeInfer(self.dialects)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = RewriteResult()
        frame, _ = self.constprop.run_analysis(mt)
        result = Walk(WrapConst(frame)).rewrite(mt.code).join(result)
        rule = Chain(
            ConstantFold(),
            Call2Invoke(),
            InlineGetField(),
            InlineGetItem(),
            DeadCodeElimination(),
            CommonSubexpressionElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)

        result = (
            Walk(
                Chain(
                    scf.unroll.PickIfElse(),
                    scf.unroll.ForLoop(),
                    scf.trim.UnusedYield(),
                )
            )
            .rewrite(mt.code)
            .join(result)
        )

        # run typeinfer again after unroll etc. because we now insert
        # a lot of new nodes, which might have more precise types
        self.typeinfer.unsafe_run(mt)
        result = (
            Walk(Chain(ilist.rewrite.ConstList2IList(), ilist.rewrite.Unroll()))
            .rewrite(mt.code)
            .join(result)
        )

        def inline_simple(node: ir.Statement):
            if isinstance(node, expr.GateFunction):
                return self.inline_gate_subroutine

            if not isinstance(node.parent_stmt, (scf.For, scf.IfElse)):
                return True  # always inline calls outside of loops and if-else

            # inside loops and if-else, only inline simple functions, i.e. functions with a single block
            if (trait := node.get_trait(ir.CallableStmtInterface)) is None:
                return False  # not a callable, don't inline to be safe
            region = trait.get_callable_region(node)
            return len(region.blocks) == 1

        result = (
            Walk(
                Inline(inline_simple),
            )
            .rewrite(mt.code)
            .join(result)
        )
        result = Walk(Fixpoint(CFGCompactify())).rewrite(mt.code).join(result)
        return result



import math
import textwrap

import cirq
import cirq.contrib
import cirq.testing
import cirq.contrib.qasm_import as qasm_import
import cirq.circuits.qasm_output as qasm_output
from pytest import mark
from bloqade import qasm2
from kirin.rewrite import walk
from bloqade.qasm2.rewrite.native_gates import (
    RydbergGateSetRewriteRule,
    one_qubit_gate_to_u3_angles,
)


def test_one_qubit_gate_to_u3_angles():
    theta = 1.1 * math.pi
    phi = 0.2 * math.pi
    lam = 1.6 * math.pi

    op = qasm_output.QasmUGate(theta / math.pi, phi / math.pi, lam / math.pi)(
        cirq.LineQubit(0)
    )

    theta, phi, lam = one_qubit_gate_to_u3_angles(op)
    op1 = qasm_output.QasmUGate(theta / math.pi, phi / math.pi, lam / math.pi)(
        cirq.LineQubit(0)
    )

    cirq.testing.assert_allclose_up_to_global_phase(
        cirq.unitary(op), cirq.unitary(op1), atol=1e-8
    )


def generator(n_tests: int):
    import numpy as np

    yield textwrap.dedent(
        """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[1];

    tdg q[0];

    """
    )

    yield textwrap.dedent(
        """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[1];

    sdg q[0];

    """
    )

    yield textwrap.dedent(
        """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[1];

    sx q[0];

    """
    )

    yield textwrap.dedent(
        """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[1];

    sxdg q[0];

    """
    )

    rgen = np.random.RandomState(128)
    for num in range(n_tests):
        # Generate a new instance:
        N = int(rgen.choice([4, 5, 6, 7, 8]))
        D = int(rgen.choice([1, 3, 5, 10, 15]))
        rho = float(rgen.random())
        circ: cirq.Circuit = cirq.testing.random_circuit(N, D, rho, random_state=rgen)
        while len(circ.all_qubits()) == 0:
            circ = cirq.testing.random_circuit(N, D, rho, random_state=rgen)
        yield circ.to_qasm()


@mark.parametrize(
    "qasm2_prog",
    generator(20),
)
def test_rewrite(qasm2_prog: str):
    @qasm2.main.add(qasm2.dialects.inline)
    def kernel():
        qasm2.inline(qasm2_prog)

    walk.Walk(RydbergGateSetRewriteRule(kernel.dialects)).rewrite(kernel.code)

    new_qasm2 = qasm2.emit.QASM2().emit_str(kernel)

    cirq_circuit = qasm_import.circuit_from_qasm(new_qasm2)
    old_circuit = qasm_import.circuit_from_qasm(qasm2_prog)

    cirq.testing.assert_allclose_up_to_global_phase(
        cirq.unitary(old_circuit), cirq.unitary(cirq_circuit), atol=1e-8
    )



from bloqade import qasm2
from bloqade.test_utils import assert_methods
from bloqade.qasm2.passes import Py2QASM, QASM2Py


def test_round_trip():
    @qasm2.main
    def test():
        q = qasm2.qreg(2)
        c = qasm2.creg(1)
        theta = qasm2.sin(1.0) / qasm2.sqrt(1.0)
        phi = qasm2.cos(1.0) * qasm2.exp(1.0)
        lam = -qasm2.tan(1.0) + qasm2.exp(1.0) - 2.0**4.0
        if c == c:
            qasm2.u(q[0], theta, phi, lam)

    test_original = test.similar()

    QASM2Py(test.dialects)(test)

    test = test.similar(qasm2.extended)

    Py2QASM(test.dialects)(test)

    test = test.similar(qasm2.main)
    assert_methods(test, test_original)
